
open Java_context_types;;
open Ast_types;;
open Ast;;
open Pre_translate;;

let sprintf = Format.sprintf

let tr_ive ive = ive.instance_variable_entry_trans_name
(*i
  if ive.instance_variable_entry_is_ambiguous
  then 
    ive.instance_variable_entry_class.class_entry_name ^ "_" ^ 
    ive.instance_variable_entry_name
  else ive.instance_variable_entry_name
i*)

let type_default_value produce_a_value t =
  match t with
    | Integer_type -> 
	let v = Why.Cte(Why.Prim_int(0)) in
	if produce_a_value then Why.make_app "Primitive_int" [v] else v
    | Float_type -> 
	let v = Why.Cte(Why.Prim_float(0.0)) in
	if produce_a_value then Why.make_app "Primitive_float" [v] else v
    | Boolean_type -> 
	let v = Why.Cte(Why.Prim_bool(false)) in
	if produce_a_value then Why.make_app "Primitive_bool" [v] else v
    | Class_type(_) -> Why.Var("Null")
    | Array_type _ ->  Why.Var("Null")
    | Interface_type _ -> assert false
    | Null_type -> assert false
;;
let has_head_term s = function 
    Why.LApp(s',_) -> s=s'
  | _ -> false 

let has_head_expr s = function 
    Why.App(Why.Var(s'),_) -> s=s'
  | _ -> false 

let arg_term = function 
    Why.LApp(_,[x]) -> x
  | _ -> assert false 

let arg_expr = function 
    Why.App(_,x) -> x
  | _ -> assert false 

let ref_modifiable = ref Why.LTrue


let coerce_to_value_term t accu = match t with 
      | Integer_type -> 
	  if has_head_term "int_val" accu then arg_term accu 
	  else Why.LApp("Primitive_int",[accu])
      | Boolean_type -> 
	  if has_head_term "bool_val" accu then arg_term accu 
	  else Why.LApp("Primitive_bool",[accu])
      | Float_type -> 
	  if has_head_term "float_val" accu then arg_term accu 
	  else Why.LApp("Primitive_float",[accu])
      | Null_type -> ill_typed "coerce_to_value_term(null)"
      | Class_type(_) 
      | Array_type(_)
      | Interface_type(_) ->  accu
	
let coerce_to_value_expr t accu = match t with 
      | Integer_type -> 
	  if has_head_expr "int_val" accu then arg_expr accu 
	  else Why.make_app "Primitive_int" [accu]
      | Boolean_type -> 
	  if has_head_expr "bool_val" accu then arg_expr accu 
	  else Why.make_app "Primitive_bool" [accu]
      | Float_type -> 
	  if has_head_expr "float_val" accu then arg_expr accu 
	  else Why.make_app "Primitive_float" [accu]
      | Null_type -> ill_typed "coerce_to_value_expr(null)"
      | Class_type(_) 
      | Array_type(_)
      | Interface_type(_) ->  accu
	
(* translation of types : to primitive type or to "Value" *)

let tr_type t =
  match t with
    | Integer_type -> Why.int_type
    | Float_type -> Why.float_type
    | Boolean_type -> Why.bool_type
    | Class_type(ce) -> 
	Why.base_type(sprintf "value (* %s *)" ce.class_entry_name)
    | Array_type t -> 
	Format.fprintf 
	  Format.str_formatter
	  "value (* %a array *)" Java_context.fprintf_java_type t;
	Why.base_type (Format.flush_str_formatter ())
    | Interface_type ie -> 	
	Why.base_type(sprintf "value (* %s *)" ie.interface_entry_name)
    | Null_type -> Why.unit_type

let rec tr_type_detailed t =
  match t with
    | Integer_type -> "int"
    | Float_type -> "float"
    | Boolean_type -> "bool"
    | Class_type(ce) -> ce.class_entry_name
    | Array_type t -> "value"
    | Interface_type _ -> unsupported_no_loc "tr_type_detailed(interface_type)"
    | Null_type -> "unit"

let rec tr_type_term t =
  match t with
    | Integer_type -> Why.LVar("IntType")
    | Float_type -> Why.LVar("FloatType")
    | Boolean_type -> Why.LVar("BoolType")
    | Class_type(ce) -> 
	Why.LApp ("ClassType", [ Why.LVar ce.class_entry_name ])
    | Array_type t1 -> 
	Why.LApp
	  ("ArrayType",
	   [tr_type_term t1])
    | Interface_type _ -> 
	unsupported_no_loc "tr_type_in_coq(interface_type)"
    | Null_type -> 
	unsupported_no_loc "tr_type_in_coq(null_type)"

let rec tr_type_expr t =
  match t with
    | Integer_type -> Why.Var("IntType")
    | Float_type -> Why.Var("FloatType")
    | Boolean_type -> Why.Var("BoolType")
    | Class_type(ce) -> 
	Why.make_app "ClassType" [Why.Var ce.class_entry_name]
    | Array_type t1 -> 
	Why.make_app "ArrayType" [tr_type_expr t1]
    | Interface_type _ -> 
	unsupported_no_loc "tr_type_expr(interface_type)"
    | Null_type -> 
	unsupported_no_loc "tr_type_expr(null_type)"

let tr_literal loc l =
  match l with
    | Int(n) -> Why.Cte(Why.Prim_int(n))
    | Float(f) -> Why.Cte(Why.Prim_float(f))
    | Bool(b) -> Why.Cte(Why.Prim_bool(b))
    | String(s) -> unsupported loc "tr_literal(string)"
    | Char(c) -> unsupported loc "tr_literal(char)"
    | Null -> Why.Var("Null")
;;



let why_one = Why.Cte(Why.Prim_int(1));;

let coercion_term t e =
  match t with
    | Integer_type -> Why.LApp("int_val",[e])
    | Boolean_type -> Why.LApp("bool_val",[e])
    | Float_type -> Why.LApp("float_val",[e])
    | Class_type(_) -> e
    | Array_type(_) -> e
    | Interface_type(_) -> e
    | Null_type -> ill_typed "coercion_term(null)"
;;


let tr_logic_bin_op loc op t1 t2 =
  let op' = 
    match op with
      | "&&" -> unsupported loc "tr_logic_bin_op(&&)"
      | "||" -> unsupported loc "tr_logic_bin_op(||)"
      | "le_int" | "ge_int" | "gt_int" | "lt_int" 
      | "le_float" | "ge_float" | "gt_float" | "lt_float" 
      | "eq_int" | "eq_float" | "ne_int" | "ne_float" ->
	  (*i unsupported loc "tr_logic_bin_op(comp)" i*)
	  op
      | "ne_value" | "eq_value" -> op
      | "add_int" | "sub_int" | "mul_int" | "div_int" | "mod_int" -> op
      | "add_float" | "sub_float" | "mul_float" | "div_float" -> op
      | "Zland" | "Zlor" | "Zlsr" | "Zlsl" | "Zasr" -> op
      | _ -> unsupported loc ("tr_logic_op(" ^ op ^ ")")
  in
  Why.LApp(op',[t1;t2])
;;

let tr_logic_un_op loc op t =
  let op' =
    match op with
    | "!" -> "negb"
    | "neg_int" -> op
    | _ -> unsupported loc ("tr_logic_un_op(" ^ op ^ ")")
  in
  Why.LApp(op',[t])
;;

let rec tr_jml_lit loc l =
  match l with
    | Int(n) -> Why.LConst(Why.Prim_int(n))
    | Bool(b) -> Why.LConst(Why.Prim_bool(b))
    | Char(_) -> unsupported loc "tr_jml_lit(char)"
    | String(_) -> unsupported loc "tr_jml_lit(string)"
    | Float(_) -> unsupported loc "tr_jml_lit(float)"
    | Null -> Why.LVar("Null")
;;

(* label is the label to refer when oldfiying *)

let this_is_result = ref false;;

let rec tr_term label oldify t =
  match t.expr_node with
    | Lit(l) -> tr_jml_lit t.expr_loc l
    | Var(s) -> ill_typed ("tr_term(var "^(snd s)^" has not been typed)")
    | Typed_var(lve) ->
	let id = lve.local_variable_entry_name in
	if id = "this" && !this_is_result
	then Why.LVar("result")
	else
	  Why.LVar(if lve.local_variable_entry_is_ref 
		   then id ^ oldify else id)
    | Bin(t1,op,t2) ->
	tr_logic_bin_op t.expr_loc op 
	  (tr_term label oldify t1) (tr_term label oldify t2)
    | Un(op,t2) -> tr_logic_un_op t.expr_loc op (tr_term label oldify t2)
    | PrePostIncrDecr(op,t2) -> 
	ill_typed "tr_term(prepostincrdecr)"
    | Instanceof(_,_) -> 
	ill_typed "tr_term(instanceof)"
    | Typed_instanceof (e, t) -> 
	Why.LApp("instanceof_bool",[tr_term label oldify e;tr_type_term t])
    | Cast (t, e) -> ill_typed "tr_term(cast)"
    | Typed_cast(t,e) -> tr_term label oldify e	(* TODO ! *)
    | Array_access (e1, e2) -> ill_typed "tr_term(array_access)"
    | Typed_array_access (ty, e1, e2) -> 
	let heap = "heap" ^ oldify in
	let acc = 	  
	  Why.LApp("access",
		   [Why.LVar(heap);
		    tr_term label oldify e1;
		    Why.LApp("Index",[tr_term label oldify e2])])
	in coercion_term ty acc
    | Array_creation _ -> unsupported t.expr_loc "tr_term(new_array)"
    | Class_instance_creation (_, _) -> 
	ill_typed "tr_term(new_object)"
    | Typed_class_instance_creation (_, _) -> 
	unsupported t.expr_loc "tr_term(new_object)"
    | Method_call _ -> ill_typed "tr_term(method_call)"
    | Typed_method_call(e,me,args) -> 
	if me.method_entry_is_pure then
	begin
	  let args = List.map (tr_term label oldify) args in
	  let args =
	    if me.method_entry_is_static
	    then args
	    else 
	      let tmp_expr = tr_term label oldify e in
	      tmp_expr::args
	  in
	  let args =
	    if me.method_entry_routine.routine_entry_effects.reads_heap
	    then
	      let heap = "heap" ^ oldify in
	      Why.LVar(heap)::args
	    else
	      args
	  in
	  let class_name =
	    match me.method_entry_class_or_interface with
	      | Class_entry ce -> ce.class_entry_name
	      | Interface_entry te -> te.interface_entry_name
	      | _ -> assert false
	  in
	  Why.LApp
	    (class_name ^ "_" ^ me.method_entry_name,args)
	end
	else
	  ill_typed ("tr_term(non pure method '" ^ me.method_entry_name ^ "')")
    | Field_access fa -> tr_field_access_term t.expr_loc label oldify fa
    | If_expr (e1, e2, e3) -> 
	Why.LApp("if_then_else",
		 [tr_term label oldify e1;
		  tr_term label oldify e2;
		  tr_term label oldify e3])
    | Assign_array (_, _, _, _) -> ill_typed "tr_term(assign_array)"
    | Typed_assign_array (_, _, _, _, _) -> 
	ill_typed "tr_term(typed_assign_array)"
    | Assign_field (_, _, _) -> 
	ill_typed "tr_term(field_assign)"	
    | Assign_local_var (_, _, _) -> 
	ill_typed "tr_term(var_assign)"
    | Assign_name (_, _, _) -> 
	ill_typed "tr_term(assign_name)"
    | Static_class _ ->  
	unsupported t.expr_loc "tr_term(static_class)"
    | Static_interface(ie) -> 
	unsupported t.expr_loc "tr_term(static_interface)"
    | This -> 
	ill_typed "tr_term(this)"
    | Result -> 
	ill_typed "tr_term(\\result)"
    | Old t' -> 
	tr_term label label t'
    | Fresh t -> 
	Why.LApp("fresh",[tr_term label oldify t])
    | Quantifier(_) -> 
	ill_typed "tr_term(quantifier)"
    | Typed_quantifier(_) -> 
	unsupported t.expr_loc "tr_term(quantifier)"

and tr_field_access_term loc label oldify f =
  match f with
    | Super_access(id) -> 
	unsupported loc "tr_field_access_term(super)"
    | Primary_access(_,_) ->
	ill_typed "tr_field_access_term(primary)"
    | Typed_constant_access(ive) ->
	let c = ive.constant_entry_interface.interface_entry_name 
		^ "_" ^ ive.constant_entry_name
	in
	Why.LVar(c)
    | Typed_primary_access(e,ive) ->
	begin
	  if ive.instance_variable_entry_is_static 
	  then
	    let c = ive.instance_variable_entry_trans_name
	    in
	    Why.LVar
	      (if not ive.instance_variable_entry_is_final 
	       then c ^ oldify else c)
	  else
	    let heap = "heap" ^ oldify in
	    let acc =
	      Why.LApp
		("access",
		 [Why.LVar(heap) ; tr_term label oldify e;
		  Why.LApp("Field",
			   [Why.LVar(tr_ive ive)])])
	    in coercion_term ive.instance_variable_entry_type acc
	end
    | Array_length(e) ->
	Why.LApp("arraylength",
		 [tr_term label oldify e])


let tr_term_variable_initializer loc i =
  match i with
    | Simple_initializer(e) -> tr_term "" "" e
    | Array_initializer _ -> 
	unsupported loc "tr_term_variable_initializer(array)"
;;


let make_bin label oldify t1 op t2 =
  Why.LPred(op,[tr_term label oldify t1; tr_term label oldify t2])

let rec tr_predicate_bin_op loc label oldify op t1 t2 =
  match op with
    | "&&" -> Why.make_and 
	(tr_assertion_rec label oldify t1) 
	(tr_assertion_rec label oldify t2)
    | "||" -> Why.make_or 
	  (tr_assertion_rec label oldify t1)
	  (tr_assertion_rec label oldify t2)
    | "==>" -> 
	Why.make_impl 
	(tr_assertion_rec label oldify t1) 
	(tr_assertion_rec label oldify t2)
    | "<==" -> 
	Why.make_impl 
	(tr_assertion_rec label oldify t2) 
	(tr_assertion_rec label oldify t1)
    | "<==>" -> 
	Why.make_equiv 
	(tr_assertion_rec label oldify t1) 
	(tr_assertion_rec label oldify t2)
    | "<=!=>" -> 
	Why.LNot(Why.make_equiv
		   (tr_assertion_rec label oldify t1) 
		   (tr_assertion_rec label oldify t2))
    | "le_int" | "ge_int" | "gt_int" | "lt_int" 
    | "le_float" | "ge_float" | "gt_float" | "lt_float"
    | "eq_int" | "eq_float" | "eq_bool" -> make_bin label oldify t1 op t2	
    | "ne_int" -> make_bin label oldify t1 "neq_int" t2
    | "ne_float" -> make_bin label oldify t1 "neq_float" t2
    | "ne_bool" -> make_bin label oldify t1 "neq_bool" t2
    | "ne_value" | "eq_value" -> make_bin label oldify t1 op t2
    | "add_int" | "sub_int" | "mul_int" | "div_int" | "mod_int" 
    | "add_float" | "sub_float" | "mul_float" | "div_float" 
    | "land" 
    | "lor" 
    | "lsr" 
    | "lsl" 
    | "asr" -> ill_typed "tr_predicate_op"
    | _ -> unsupported loc ("tr_predicate_bin_op(" ^ op ^ ")")

and tr_assertion_rec label oldify a =   
  match a.expr_node with
    | Bin(t1,op,t2) ->
	tr_predicate_bin_op a.expr_loc label oldify op t1 t2
    | Un("!",t2) ->
	Why.LNot(tr_assertion_rec label oldify t2)
    | If_expr(b,t1,t2) ->
	Why.LIf(tr_term label oldify b, 
		tr_assertion_rec label oldify t1,
		tr_assertion_rec label oldify t2)
    | Lit(Bool(true)) -> Why.LTrue
    | Lit(Bool(false)) -> Why.LFalse
    | Typed_instanceof (e, t) -> 
	Why.LPred("instanceof",[tr_term label oldify e;tr_type_term t])
(*i
    | Typed_var(lve) ->
	let id = lve.local_variable_entry_name in
	let v = 
	  Why.LVar(if oldify && lve.local_variable_entry_is_ref
		   then id ^"@init" else id)
	in
	Why.Term(Why.LApp("eq_bool",[v;Why.LConst(Why.Prim_bool(true))]))
i*)
	
    | Typed_method_call(e,me,args) when me.method_entry_is_predicate ->
	begin
	if me.method_entry_is_pure then
	begin
	  let args = List.map (tr_term label oldify) args in
	  let args =
	    if me.method_entry_is_static
	    then args
	    else 
	      let tmp_expr = tr_term label oldify e in
	      tmp_expr::args
	  in
	  let args =
	    if me.method_entry_routine.routine_entry_effects.reads_heap
	    then
	      let heap = "heap" ^ oldify in
	      Why.LVar(heap)::args
	    else
	      args
	  in
	  let class_name =
	    match me.method_entry_class_or_interface with
	      | Class_entry ce -> ce.class_entry_name
	      | Interface_entry te -> te.interface_entry_name
	      | _ -> assert false
	  in
	  Why.LPred
	    (class_name ^ "_" ^ me.method_entry_name,args)
	end
	else
	  ill_typed "tr_assertion(non pure method)"
(*
	  match e.expr_node with
	    | Static_class(ce) ->
		let args =
		  if me.method_entry_effects.reads_heap then
		    let heap = "heap" ^ oldify in
		    Why.LVar(heap)::
		    List.map (tr_term label oldify) args
		  else
		    List.map (tr_term label oldify) args
		in
		Why.LPred(ce.class_entry_name ^ "_" ^ me.method_entry_name,
			  args)
	    | Typed_var(lve) ->
		begin
		  match lve.local_variable_entry_type with
		    | Class_type(ce) ->
			let args =
			  if me.method_entry_effects.reads_heap then
			    let heap = "heap" ^ oldify in
			    Why.LVar(heap)::
			    List.map (tr_term label oldify) (e::args)
			  else
			    List.map (tr_term label oldify) (e::args)
			in
			Why.LPred
			  (ce.class_entry_name ^ "_" ^ me.method_entry_name,
			   args)
		    | _ -> ill_typed "tr_term(method_call(typed_var))"
		end
	    | Typed_cast(tt,{expr_node = Typed_var(lve)}) ->
		begin
		  match tt with
		    | Class_type(_) | Interface_type(_) ->
			let args =
			  if me.method_entry_effects.reads_heap then
			    let heap = "heap" ^ oldify in
			    Why.LVar(heap)::
			    List.map (tr_term label oldify) (e::args)
			  else
			    List.map (tr_term label oldify) (e::args)
			in
			Why.LPred
			  ((type_entry_name tt)^ "_" ^ me.method_entry_name,
			   args)
		    | _ ->
 			unsupported e.expr_loc "tr_term(method_call(cast))"
		end
	    | _ -> 
		Format.fprintf Config.log 
		"unsupported: tr_assertion(method_call(%a))@." 
		Ast.fprintf_expr e;
		unsupported e.expr_loc "tr_term(method_call(novar))"	
*)
	end
    | Typed_quantifier(Forall,idl,e1,e2) -> 
	List.fold_right
	  (fun lve acc ->
	     let ty = tr_type_detailed lve.local_variable_entry_type in
	     Why.LForall(lve.local_variable_entry_name,ty,acc))
	  idl
	  (Why.LImpl(tr_assertion_rec label oldify e1,
		     tr_assertion_rec label oldify e2))
    | Typed_quantifier(Exists,idl,e1,e2) -> 
	List.fold_right
	  (fun lve acc ->
	     let ty = tr_type_detailed lve.local_variable_entry_type in
	     Why.LExists(lve.local_variable_entry_name,ty,acc))
	  idl
	  (Why.make_and (tr_assertion_rec label oldify e1)
		     (tr_assertion_rec label oldify e2))
    | Old b -> tr_assertion_rec label label b
    | Fresh b -> Why.LPred("fresh",
			   [Why.LVar("heap" ^ label);
			    tr_term label label b])
    | _ -> 
	Why.LPred("eq_bool",[tr_term label oldify a;
			     Why.LConst(Why.Prim_bool(true))])

let tr_assertion label = tr_assertion_rec label ""


let tr_bin_op loc s =
  let s' = 
    match s with
      | "add_int" | "sub_int" | "mul_int" | "div_int" | "mod_int" -> s
      | "add_float" | "sub_float" | "mul_float" | "div_float" -> s
      | "le_int" | "ge_int" | "gt_int" | "lt_int" -> s
      | "le_float" | "ge_float" | "gt_float" | "lt_float" -> s
      | "||" | "&&" -> s
      | "Zland" | "Zlor" | "Zlsr" | "Zlsl" | "Zasr" -> s
      | "andb" | "orb" -> s
      | "eq_int" -> s
      | "eq_float" -> s
      | "eq_bool" -> s
      | "ne_int" -> "neq_int"
      | "ne_float" -> "neq_float"
      | "ne_bool" -> "neq_bool"
      | "ne_value" | "eq_value" -> s
      | _ -> unsupported loc ("tr_bin_op(" ^ s ^ ")")
  in
  Why.Var(s')

let tr_un_op loc s =
  let s' = 
   match s with
    | "neg_int" -> s
    | "neg_float" -> s
    | "float_of_int" -> s
    | "!" -> "negb"
    | _ -> unsupported loc ("tr_un_op(" ^ s ^ ")")
  in
  Why.Var(s')


(*

 coercion

*)

let coercion_expr t e =
  match t with
    | Integer_type -> Why.make_app "int_val" [e]
    | Boolean_type -> Why.make_app "bool_val" [e]
    | Float_type -> Why.make_app "float_val" [e]
    | Class_type(_) -> e
    | Array_type(_) -> e
    | Interface_type(_) -> e
    | Null_type -> ill_typed "coercion_expr(null)"
;;

(*

  auxiliary functions to detect purity of expressions, and decide when
to introduce intermediaite variables or not.

*)

let is_pure e =
  let eff = Effects.expr e in 
  (not eff.writes_heap) && (Inttagset.is_empty eff.writes) && (not eff.throws) && (not eff.breaks)
  && 
  (match e.expr_node with 
     | Method_call _ -> assert false
     | Typed_method_call _ -> false
     | _ -> true)
;;

let build_why_app e args =
  let rec build e args =
    match args with
      | [] -> e
      | p::l -> build (Why.App(e,p)) l
  in
  match args with
    | [] -> Why.App(e,Why.Void)
    | _ -> build e args
	  

let tempindex_count = ref(-1);;

let temp_index () =
  incr tempindex_count;
  if !tempindex_count = 0
  then "krak_index"
  else "krak_index" ^ (string_of_int !tempindex_count)
;;

(* Build a Let only if the expression is complex 
    [v : string,f : Why.expr -> Why.term -> Why.expr; e: Why.expr] 
*)

let why_let_var is_pure v f e = match e with 
  | (Why.Cte n) -> f e (Why.LConst n)
  | (Why.Var x) -> f e (Why.LVar x) 
  | (Why.Deref x) when is_pure -> f e (Why.LVar x) 
  | _ -> Why.Let(v,e,f (Why.Var v) (Why.LVar v))

let why_let_var_ref = why_let_var true
(*i
v f e = match e with 
    (Why.Var x) -> f e (Why.LVar x)
  | (Why.Cte n) -> f e (Why.LConst n)
  | (Why.Deref x) -> f e (Why.LVar x) 
  | _ -> Why.Let(v,e,f (Why.Var v) (Why.LVar v))
i*)


let tempvar_count = ref 0;;

let build_complex_app e args =
  let rec build n e args =
    match args with
      | [] -> e
      | [p] -> Why.App(e,p)
      | ((Why.Var _) | (Why.Cte _) as p)::l ->
	  build n (Why.App(e,p)) l
      | p::l ->
	  incr tempvar_count;
	  let v = "krakv"^(string_of_int !tempvar_count) in
	  Why.Let(v,p,
	    build (succ n) (Why.App(e,Why.Var(v))) l)
  in
  match args with
    | [] -> Why.App(e,Why.Void)
    | _ -> build 1 e args

let rec build_app1 e args =
  if List.for_all is_pure args then
    List.fold_left (fun acc x -> Why.App(acc,tr_expr x)) e args
  else
    build_complex_app e (List.map tr_expr args)

and build_app e args =
  match args with
    | [] -> Why.App(e,Why.Void)
    | _ -> build_app1 e args




(* translation of cast (pas au point !)*)

and tr_cast t e = 
  match t with
(*i
    | Base_type(s) -> tr_expr e (* TODO ! *)	
    | Type_name(qi) ->
	begin
	  match qi with
	    | [(loc,id)] -> 
		let result_type = Why.Var(id) (* correct ?? *) in
		Why.make_app "cast" [result_type;tr_expr e]
	    | _ ->unsupported e.expr_loc "tr_cast(longname)"
	end
    | Array_type_expr(t) -> unsupported e.expr_loc "tr_cast(array)"
i*)
  | Integer_type
  | Float_type
  | Boolean_type -> tr_expr e (* TODO ! *)	
  | Null_type -> assert false
(*i
  | Class_type(te) ->
      let result_type = Why.Var(te.class_entry_name) in
      Why.make_app "cast" [result_type;tr_expr e]      
  | Interface_type(te) ->
      let result_type = Why.Var(te.interface_entry_name) in
      Why.make_app "cast" [result_type;tr_expr e]      
i*)
(*i
  | _ ->
      let result_type = tr_type_expr t in
      Why.make_app "cast" [result_type;tr_expr e]      
i*)
  | _ ->
      let result_type = tr_type_term t in
      why_let_var (is_pure e)
	"krak_cast"
        (fun whye loge ->
	   Why.make_pre
	     (Why.LPred("instanceof",[loge;result_type]))
	     whye)
	(tr_expr e)


and tr_primary_access e ive = 
  why_let_var_ref (* we allow deref here because (Field ive) is always pure *)
    "krak_access"
    (fun whye loge -> 
       let pre =
	 Why.LPred("is_valid_cell",[loge;
				    Why.LApp("Field",
					     [Why.LVar(tr_ive ive)])])
       in
       let acc = 
	 Why.make_app "access"
	   [Why.Deref("heap") ;
	    whye;
	    Why.make_app "Field"
	      [Why.Var(tr_ive ive)]]
       in
       Why.make_pre pre (coercion_expr ive.instance_variable_entry_type acc))
    (tr_expr e)

and tr_field_access loc f =
  match f with
    | Super_access(id) -> unsupported loc "tr_field_access(super)"
    | Primary_access(e,id) -> ill_typed "tr_field_access(primary)"
    | Typed_primary_access(e,ive) ->
	begin
	  if ive.instance_variable_entry_is_static
	  then
	    if ive.instance_variable_entry_is_final
	    then Why.Var(ive.instance_variable_entry_trans_name) 
	    else Why.Deref(ive.instance_variable_entry_trans_name) 
	  else
	    tr_primary_access e ive
	end
    | Typed_constant_access(const) ->
	let c = const.constant_entry_interface.interface_entry_name in
	Why.Var(c ^ "_" ^ const.constant_entry_name) 
    | Array_length(e) ->
	why_let_var_ref 
	  "krak_access"
	  (fun whye loge ->
	     let acc =
	       Why.make_app "arraylength" [whye]
	     in
	     match loge with
	       | Why.LVar "this" -> acc
	       | _ -> 
		   let pre =
		     Why.LPred("neqv",[loge;Why.LVar("Null")])
		   in
		   Why.make_pre pre acc)
	  (tr_expr e)

and tr_pre_post_incr_decr op e =
  match e.expr_node with
    | Var(x) -> ill_typed "tr_pre_post_incr_decr(var)"
    | This -> ill_typed "tr_pre_post_incr_decr(this)"
    | Typed_var(lve) -> 
	let v = lve.local_variable_entry_name in
	Why.Assign
	  (v,build_why_app (tr_bin_op e.expr_loc op) [Why.Deref(v);why_one]) 
    | Field_access(fa) -> 
	begin
	  match fa with
	    | Array_length _ -> ill_typed "tr_pre_post_incr_decr(length)"
	    | Typed_constant_access _ -> ill_typed "tr_pre_post_incr_decr(constant)"
	    | Primary_access (_, _) -> 
		ill_typed "tr_pre_post_incr_decr(primary)"
	    | Super_access _ -> 
		unsupported e.expr_loc "tr_pre_post_incr_decr(super)"
	    | Typed_primary_access(f,ive) ->
		let tr = tr_expr f in
		tr_field_assignment_op e.expr_loc tr ive op why_one
	end
    | _ -> unsupported e.expr_loc "tr_pre_post_incr_decr()"

and tr_post_incr_decr op e =
  unsupported e.expr_loc "tr_post_incr_decr"

and tr_pre_incr_decr op e =
  unsupported e.expr_loc "tr_pre_incr_decr"

and tr_field_assignment loc f ive is_pure accu = 
  if ive.instance_variable_entry_is_static
  then
    Why.Assign(ive.instance_variable_entry_trans_name, accu)
  else
    let t = ive.instance_variable_entry_type in
    why_let_var is_pure "krak_update"
      (fun whyf logf -> 
	 why_let_var is_pure "krak_acc"
	   (fun whyaccu logaccu ->
	    let pre =
	      Why.LPred
		("is_valid_update",
		 [Why.LVar("heap");
		  logf;
		  Why.LApp
		    ("Field",
		     [Why.LVar(tr_ive ive)]);
		  (coerce_to_value_term t logaccu)])
	    in 
	    Why.make_pre
	      pre
	      (Why.Assign
		 ("heap", 
		  Why.make_app "update"
		    [Why.Deref("heap");
		     whyf;
		     Why.make_app "Field"
		       [Why.Var(tr_ive ive)];
		     (coerce_to_value_expr t whyaccu)])))
	   accu)
      (tr_expr f)
    
and tr_field_assignment_op loc f ive op accu = 
  if ive.instance_variable_entry_is_static
  then
    let c = ive.instance_variable_entry_trans_name
    in
    Why.Assign(c, build_why_app (tr_bin_op loc op) [Why.Deref(c); accu])
  else
    (* on traduit f.ive op= accu par
       let krak_update = f in
       let kra_accu = (op (access !heap krak_update (Field ive)) accu) in
       { is_valid_update heap krak_update (Field ive) accu }
       heap := update !heap krak_update (Field ive) accu
       {}
       *)
    let lfield = 
      Why.LApp("Field",[Why.LVar(tr_ive ive)])
    and field = 
      Why.make_app "Field" [Why.Var(tr_ive ive)]
    in
    why_let_var false "krak_update"
      (fun whyf logf -> 
	 let pre_access =
	   Why.LPred("is_valid_cell",[logf;lfield])
	 in
	 let acc = 
	   Why.make_app "access"
	     [Why.Deref("heap"); whyf; field]
	 in
	 let access =
	   Why.make_pre
	     pre_access
	     (coercion_expr ive.instance_variable_entry_type acc)
	 in
	 why_let_var false "krak_acc"
	   (fun whyaccu logaccu ->
	      let t = ive.instance_variable_entry_type 
	      in 
	      let pre =
		Why.LPred
		     ("is_valid_update",
		      [Why.LVar("heap");
		       logf;lfield;coerce_to_value_term t logaccu])
	      in 
	      Why.make_pre
		pre
		(Why.Assign
		   ("heap", 
		    Why.make_app "update" 
		      [Why.Deref("heap");
		       whyf;
		       field;
		       coerce_to_value_expr t 
			 whyaccu])))
	   (build_why_app (tr_bin_op loc op) [access;accu]))
    f
    
and tr_array_assignment t e1 e2 accu_is_pure accu = 
  why_let_var (accu_is_pure && (is_pure e2)) "krak_update" 
    (fun whye1 loge1 ->
       why_let_var accu_is_pure (temp_index ())
	 (fun whye2 loge2 ->
	    why_let_var true "krak_acc"
	      (fun whyaccu logaccu -> 
		 let pre =
		   Why.LPred
		     ("is_valid_update",
		      [Why.LVar("heap");
		       loge1; 
		       Why.LApp("Index",[loge2]);
		       coerce_to_value_term t logaccu])
		 in         
		 Why.make_pre
		   pre
		   (Why.Assign
		      ("heap", 
		       Why.make_app "update" 
			 [Why.Deref("heap");
			  whye1;
			  Why.make_app "Index" [whye2];
			  coerce_to_value_expr t whyaccu])))
	      accu)
	 (tr_expr e2))
    (tr_expr e1)
    
and tr_method_call e me args =
      let args =
	if me.method_entry_is_static
	then args
	else e::args
      in
      let class_name =
	match me.method_entry_class_or_interface with
	  | Class_entry ce -> ce.class_entry_name
	  | Interface_entry te -> te.interface_entry_name
	  | _ -> assert false
      in
      build_app
	(Why.Var
	   (class_name ^ "_" ^ 
	    me.method_entry_name ^ "_parameter"))
	args

(*
  match e.expr_node with
    | Var(x) -> ill_typed "tr_method_call(var)"
    | This -> ill_typed "tr_method_call(this)"
    | Typed_var(lve) -> 
	begin
	  let v = lve.local_variable_entry_name in
	  match me.method_entry_class_or_interface with
            | Class_entry(ce) ->
		let this = 
		  if lve.local_variable_entry_is_ref
		  then Why.Deref(v) else Why.Var(v)
		in
		build_app1
		  (Why.App
		     (Why.Var
			(ce.class_entry_name ^ "_" ^ 
			 me.method_entry_name ^ "_parameter"),
			this))
		  args
	    | _ -> unsupported e.expr_loc "tr_method_call(var)"
	end
    | Static_class(ce) ->
	build_app
	  (Why.Var(ce.class_entry_name ^ "_" ^ me.method_entry_name ^ "_parameter")) 
	  args
    | _ -> unsupported e.expr_loc "tr_method_call()"
*)

and tr_expr e =
  match e.expr_node with
    | Lit(l) -> tr_literal e.expr_loc l
    | Bin(e1,op,e2) ->
	begin
	  match op with
	    | "&&" -> Why.And(tr_expr e1,tr_expr e2)
	    | "||" -> Why.Or(tr_expr e1,tr_expr e2)
	    | _ ->
		build_app1 (tr_bin_op e.expr_loc op) [e1;e2]
	end
    | Un(op,e1) ->
	build_app1 (tr_un_op e.expr_loc op) [e1]
    | PrePostIncrDecr(op,e1) ->
	begin
	  match op with
	    | `PREINCR -> 
		tr_pre_incr_decr "add_int" e1
	    | `POSTINCR -> 
		tr_post_incr_decr "add_int" e1
	    | `PREDECR ->
		tr_pre_incr_decr "sub_int" e1
	    | `POSTDECR -> 
		tr_post_incr_decr "sub_int" e1 
	end
    | Var(_) -> ill_typed "tr_expr(var)"
    | Typed_var(lve) -> 
	if lve.local_variable_entry_is_ref
	then Why.Deref(lve.local_variable_entry_name)
	else Why.Var(lve.local_variable_entry_name)
    | Static_class(_) -> unsupported e.expr_loc "tr_expr(static_class)"
    | Static_interface(_) -> 
	Format.fprintf Config.log
	"e = %a@." Ast.fprintf_expr e;
        unsupported e.expr_loc "tr_expr(static_interface)"
    | This -> ill_typed "tr_expr(this)" 
    | Assign_name(_,_,_) -> ill_typed "tr_expr(assign_name)"
    | Assign_local_var(lve,op,e1) ->
	begin
	  match op with
	    | "=" -> unsupported e.expr_loc "tr_expr(=)"
	    | _ -> unsupported e.expr_loc ("tr_expr("^op^")")
	end
    | Assign_field(fa,op,e1) ->
	begin
	  match fa with
	    | Super_access(id) -> 
		unsupported e.expr_loc "tr_expr(assign_field_super)"
	    | Primary_access(f,_) ->
		ill_typed "tr_expr(assign_field_primary)"
	    | Typed_constant_access(id) -> 
		ill_typed "tr_expr(assign_constant)"
	    | Typed_primary_access(f,ive) ->
		unsupported e.expr_loc 
		  (* WRONG : should return the value assigned *)
		  "tr_expr(assign_field_primary)"
		  (* Why.Let("temp_for_update",
		     tr_expr f,
		     Why.make_app "update_" ^ id), 
		     Why.Var("temp_for_update")),
		     tr_expr e)) *)
	    | Array_length(e2) -> 
		ill_typed "tr_expr(assign_field_array_length)"
	end
    | Assign_array(e1,e2,op,e3) -> ill_typed "tr_expr(assign_array)"
    | Typed_assign_array(t,e1,e2,op,e3) -> 
	unsupported e.expr_loc "tr_expr(array_assign)"
    | If_expr(e1,e2,e3) -> 
	Why.If(tr_expr e1,tr_expr e2,tr_expr e3)
    | Field_access(fa) -> tr_field_access e.expr_loc fa
    | Method_call _ -> ill_typed "tr_expr(method_call)"
    | Typed_method_call(e,me,args) ->
	tr_method_call e me args
(*
	  (List.combine 
	     (List.map 
		(fun (_,t) -> 
		   match t with
		     | Local_variable_entry le -> 
			 le.local_variable_entry_type
		     | _ -> assert false) 
		me.method_entry_parameters)
	     args)
*)
    | Array_creation(a) -> tr_array_creation e.expr_loc a
    | Array_access(_) -> ill_typed "tr_expr(array_access)"
    | Typed_array_access(t,e1,e2) -> 	
	if is_pure e1 && is_pure e2 then
	  let pre = 
	    Why.LPred
	      ("is_valid_cell",
	       [tr_term "" "" e1;
		Why.LApp("Index",[tr_term "" "" e2])])
	  and acc = 
	    Why.make_app "access"
	      [Why.Deref("heap");
	       tr_expr e1;
	       Why.make_app "Index" [tr_expr e2]]
	  in Why.make_pre pre (coercion_expr t acc)
	else
	why_let_var (is_pure e1) "krak_access"
	  (fun whye1 loge1 ->
	     why_let_var true (temp_index ())
	       (fun whye2 loge2 ->
		  let pre = 
		    Why.LPred
		      ("is_valid_cell",
		       [loge1;Why.LApp("Index",[loge2])])
		  and acc = 
		    Why.make_app "access"
		      [Why.Deref("heap");
		       whye1;
		       Why.make_app "Index" [whye2]]
		  in Why.make_pre pre (coercion_expr t acc))
	       (tr_expr e2))
	  (tr_expr e1)
    | Cast(t,e) -> ill_typed "tr_expr(cast)"
    | Typed_cast(t,e) -> tr_cast t e
    | Class_instance_creation(_,_) -> 
	ill_typed "tr_expr(new class)"
    | Typed_class_instance_creation(ce,args) -> 
	(* vieille version :
           let new_val = (new_obj heap classid) in
           heap := (store_new_ref heap new_val);
           (constr_classId new_val args)
           nouvelle version:
             (new_classId args)
        *)
(*i
	let new_val = "krak_new" in
	Why.Let
	  (new_val,
	   Why.make_app "new_obj"
	     [Why.Deref("heap");
	      Why.Var(ce.constructor_entry_class.class_entry_name)],
	   Why.Block
	     ([Why.Assign
		 ("heap",
		  Why.make_app "store_new_ref"
		    [Why.Deref("heap");
		     Why.Var(new_val)]);
	       build_why_app
		 (Why.Var(ce.constructor_entry_trans_name ^ "_parameter"))
		 (Why.Var(new_val)::(List.map tr_expr args));
	       Why.Var(new_val)]))
i*)
	build_why_app
	  (Why.Var(ce.constructor_entry_trans_name ^ "_parameter"))
		 (List.map tr_expr args);
    | Instanceof(e,t) -> ill_typed "(instanceof)"
    | Typed_instanceof(e1,t) -> 
	unsupported e.expr_loc "tr_expr(instanceof)"
	  (*
	    Why.make_app "instanceof" [tr_expr e;tr_type_in_coq t]
	  *)
    | Result -> assert false (* should never happen *)
    | Old(_) -> assert false (* should never happen *)
    | Fresh(_) -> assert false (* should never happen *)
    | Quantifier(_,_,_,_,_) -> assert false (* should never happen *)
    | Typed_quantifier(_,_,_,_) -> assert false (* should never happen *)

and tr_array_creation loc a =
  match a with
    | Explicit_array_creation(e,Typed_implicit_array_creation(t)) ->
	(*  let krak_length = e in
	    let krak_new = new_array heap krak_length t in	      
	    heap := (store_new_ref heap krak_new);
            krak_new
	*)
	(*i
	let pre =
	  Why.LPred("le_int",
		    [Why.LConst(Why.Prim_int(0));Why.LVar("krak_length")])
	in
	Why.Let
	  ("krak_length",
	   tr_expr e,
	   Why.Let
	     ("krak_new", 
	      Why.make_pre pre
		(Why.make_app "new_array"
		   [Why.Deref("heap");
		    Why.Var("krak_length");
		    tr_type_expr t]),
		Why.Block
		 [Why.Assign
		    ("heap",
		     Why.make_app "store_new_ref"
		       [Why.Deref("heap");
			Why.Var("krak_new")]);
		  Why.Var("krak_new")]))	      
	  i*)
	  (*
	    (alloc_new_array t e)	      
	  *)
	Why.make_app
	  "alloc_new_array"
	  [tr_type_expr t;tr_expr e]
	    
	  
    | Explicit_array_creation(e,a) ->
	unsupported loc "tr_array_creation(multidimensional)";
	(*  il faudrait
	    let length = e in
	    let new_val = new_array heap length type_des_cellules in	      
	    heap := (store_new_ref heap new_val);
	    for i=0 to length-1 do heap := update heap new_val i a
	*)
	(*i
	Why.Let
	  ("length",
	   tr_expr e,
	   Why.Let
	     ("new_val", 
	      (Why.make_app "new_array"
	  [Why.Deref("heap")),
		  Why.Var("length")],
	      Why.Let
		("new_vector", 
		 (Why.make_app
		    (Why.App
		       (Why.Var("new_array_map"),
			Why.Var("length")),
		       tr_array_creation loc a)),
		 (Why.Block
		    [Why.Assign
		       ("heap",
			Why.make_app
			  (Why.App
			     (Why.App(Why.Var("store_add_value"),
				      Why.Deref("heap")),
			      Why.Var("new_val")),
			     Why.Var("new_vector")));
		     Why.Var("new_val")]))))
	  i*)
    | Implicit_array_creation(_) -> ill_typed "tr_array_creation(implicit)"
    | Typed_implicit_array_creation(t) -> type_default_value true t
	  

let tr_variable_id loc id =
  match id with
    | Simple_id s -> snd s
    | Array_id _ -> unsupported loc "tr_variable_id(array)"
;;

let tr_variable_initializer loc i =
  match i with
    | Simple_initializer(e) -> tr_expr e
    | Array_initializer _ -> unsupported loc "tr_variable_initializer(array)"
;;

let tr_local_var_declarator loc t v accu =
  let init =
    match v.variable_initializer with
      | None -> type_default_value false t
      | Some e -> tr_variable_initializer loc e
  in
  Why.Let_ref(tr_variable_id loc v.variable_id,init,accu)

;;

let tr_local_var_declaration loc lve init acc =
  let init =
    match init with
      | None -> type_default_value false lve.local_variable_entry_type
      | Some e -> tr_variable_initializer loc e
  in
  Why.Let_ref(lve.local_variable_entry_name,init,acc)
;;

(*

  [abrupt_termination s] is [true] if [s] always terminate abruptly,
  by a return, a break, a continue or a throw statement. It returns
  [false] if it always terminate nromally, and raises [Not_found] if
  both are possible.

*)


let rec abrupt_termination s =
  match s with
    | Skip | Expr(_) | Typed_expr(_) -> false
    | Var_decl(v) -> false
    | Typed_var_decl(v,_,i) -> abrupt_termination i
    | Return(e) -> true
    | Break(l) -> true
    | Continue(l) -> true
    | Label(l,s) -> abrupt_termination s
    | If_statement(e,s1,s2) ->
	let c1=abrupt_termination s1
	and c2=abrupt_termination s2
	in if c1=c2 then c1 else raise Not_found
    | While(loop_inv,e,s) -> abrupt_termination s
    | For(_,_,e,el2,s) 
    | For_decl(_,_,e,el2,s) 
    | For_typed_decl(_,_,_,e,el2,s) -> 
	abrupt_termination s
    | Switch(e,cases) -> 
	List.for_all (fun (_,b) -> block_abrupt_termination b) cases
    | Block(s) -> block_abrupt_termination s
    | Synchronized (e, _) -> 
	unsupported e.expr_loc "abrupt_termination(synchronized)"
    | Try (_, _, _) -> 
	unsupported_no_loc "abrupt_termination(try)"
    | Do (inv,s,_) -> abrupt_termination s
    | Throw(_) ->  true
    | Jml_assert _ -> false
    | Annotated _ -> assert false
    | Typed_annotated(_,st) -> abrupt_termination st
	
and block_abrupt_termination s =
  match s with
    | [] -> false
    | i::l ->
	if abrupt_termination i
	then true
	else block_abrupt_termination l
;;


let rec chop_last_elem l =
  match l with
    | [] -> assert false
    | [x] -> [],x
    | h::l -> let r,x = chop_last_elem l in (h::r),x
;;

let rec make_case_expr l =
  match l with
    | [] -> assert false
    | [Case c] ->
	Why.make_app "eq_int" [Why.Var "temp_switch";
			       tr_expr c]
    | (Case c)::l -> 
	Why.Or
	  (Why.make_app "eq_int" [Why.Var "temp_switch";
				  tr_expr c],
	   make_case_expr l)
    | _ -> assert false
;;
		  
let rec well_formed_switch cases =
  match cases with
    | [] -> assert false
    | (case_list,b)::l ->	
	if List.mem Default case_list 
	then
	  begin
	    if l <> [] then Format.printf "WARNING : unreachable statement";
  	    ([],b)
	  end
	else
	  let e = make_case_expr case_list in
	  let (hb,tb) = chop_last_elem b in
	  if l=[] then
	    if tb = Break("") 
	    then ([(e,hb)],[]) 
	    else ([(e,b)],[])
	  else
	    if tb = Break("") 
	    then 
	      let (l',b') = well_formed_switch l in ((e,hb)::l',b')
	    else 
	      try
		if abrupt_termination tb
		then 
		  let (l',b') = well_formed_switch l in ((e,b)::l',b')
		else
		  raise Exit
	      with
		  Not_found -> raise Exit
;;	      




let tr_signals_clause s =
  match s with
    | None -> None
    | Some(t,var,e) ->
	assert (match var with 
		  | None -> Format.eprintf "None!@.";false
		  | Some v -> 
		      (v.local_variable_entry_name = "result") ||
		      (Format.eprintf "Some %s@." v.local_variable_entry_name;
		       false));
	Some(tr_type_detailed t, tr_assertion "@" e)
;;

let make_inter_loc l1 l2 =
  match l1,l2 with
    | (Why.LVar("nothing_loc"),_) -> l2
    | (_,Why.LVar("nothing_loc")) -> l1
    | _,_ -> Why.LApp("inter_loc",[l1;l2])
;;

let tr_assignable_base base loc oldify (assign : Ast_types.assignable_clause) =
  assert (assign <> []);
  let modif =
    List.fold_left
      (fun accu (clause,guard) ->
	 match guard with
	   | None -> 
	       let tr_clause =
		 match clause with
		   | Assignable_everything ->
		       Why.LVar("everything_loc")
		   | Assignable_nothing ->
		       Why.LVar("nothing_loc")
		   | Assignable_set_ref _ -> assert false
		   | Typed_assignable_set_ref(_,e,[sr]) ->
		       begin
			 match e.expr_node with
			   | Static_class(ce) -> Why.LVar("nothing_loc")
			   | _ ->
			       let v = tr_term oldify oldify e in 
			       match sr with 
				 | Set_field(f) -> 
				     Why.LApp
				       ("access_loc",
					[v;Why.LApp("Field",[Why.LVar(tr_ive f)])])
				 | Set_array_index(e) ->           
				     Why.LApp
				       ("access_loc",[v; Why.LApp("Index",[tr_term "@" "@" e])])
				 | Set_array_interval(e1,e2) ->
				     Why.LApp
				       ("array_sub_loc",
					[v;tr_term oldify oldify e1;
					 tr_term oldify oldify e2])
				 | Set_array ->                      
				     Why.LApp("allcells_loc",[v])
				 | Set_fieldraw(_) -> assert false
		       end
		   | Typed_assignable_set_ref(_,e,[]) -> assert false
		   | Typed_assignable_set_ref(_,e,lv) -> 
		       unsupported loc "tr_assignable(set_ref_not_empty)"
		   | Assignable_all_fields(e) ->                      
		       Why.LApp("allcells_loc",[tr_term oldify oldify e])
			 
	       in
		 make_inter_loc accu tr_clause
	   | Some e -> unsupported loc "tr_assignable(if)"
      )
      base
      assign
  in
  Why.LPred("modifiable",[Why.LVar ("heap" ^ oldify); 
			  Why.LVar "heap";modif])
;;

let tr_loop_inv label loop_spec loc =
  let inv = loop_spec.loop_invariant
  and modif = 
    match loop_spec.typed_loop_modifies with
      | None -> !ref_modifiable
      | Some clause ->
	  (Why.make_and 
	     (tr_assignable_base (Why.LVar("nothing_loc")) loc 
		("@" ^ label) clause) 
	     !ref_modifiable)
  and dec = loop_spec.loop_decreases
  in

  let i = Why.make_and (tr_assertion "@init" inv) modif    
  and d = tr_term "@init" "" dec
  in (i,d) 

let tr_assign_op loc op =
match op with
  | "add_int" | "sub_int" | "mul_int" | "div_int" | "mod_int" -> op
  | "add_float" | "sub_float" | "mul_float" | "div_float" -> op
  | "land" -> "Zland"
  | "lor" -> "Zlor"
  | "lsr" -> "Zlsr"
  | "lsl" -> "Zlsl"
  | "asr" -> "Zasr"
  | _ -> unsupported loc ("tr_assign_op("^op^")")
;;

let label_counter = ref 0;;

let rec tr_statement s = 
  tempvar_count := 0;
  tempindex_count := -1;
  match s with
    | Skip -> Why.Void
    | Expr(e) -> ill_typed "tr_statement(expr)"
    | Typed_expr(t,e) -> 
	begin
	  match e.expr_node with
	    | Lit(_) -> ill_typed "tr_statement(lit)"
	    | Bin(_,_,_) ->  ill_typed "tr_statement(binary)"
	    | Un(op,e) -> ill_typed "tr_statement(unary)"
	    | PrePostIncrDecr(op,e) ->
		begin
		  match op with
		    | `PREINCR | `POSTINCR -> 
			tr_pre_post_incr_decr "add_int" e
		    | `PREDECR | `POSTDECR -> 
			tr_pre_post_incr_decr "sub_int" e 
		end
	    | Var(_) -> ill_typed "tr_statement(var)"
	    | Typed_var(lve) -> ill_typed "tr_statement(typed_var)"
	    | Static_class(_) -> assert false (* ????????? *)
	    | Static_interface(_) -> assert false (* ????????? *)
	    | This -> ill_typed "tr_statement(this)"
	    | Assign_name(_,_,_) -> ill_typed "tr_statement(assign_name)"
	    | Assign_local_var(lve,op,e1) ->
		begin
		  match op with
		    | "=" ->
			Why.Assign(lve.local_variable_entry_name,tr_expr e1)
		    | _ ->
			Why.Assign(lve.local_variable_entry_name,
				   Why.make_app (tr_assign_op e1.expr_loc op)
				     [Why.Deref(lve.local_variable_entry_name);
				      tr_expr e1])
		end
	    | Assign_field(fa,op,e1) ->
		begin
		  match fa with
		    | Super_access(id) -> 
			unsupported e.expr_loc "assign_field(super)"
		    | Primary_access(f,(loc,id)) ->	
			ill_typed "assign_field(primary)"
		    | Typed_constant_access _ ->
			ill_typed "assign_field(constant)"			
		    | Typed_primary_access(f,ive) ->	
			begin
			  match op with
			    | "=" ->
				tr_field_assignment 
				  e.expr_loc f ive (is_pure e1) (tr_expr e1)
			    | _ -> 
				tr_field_assignment_op
				  e.expr_loc (tr_expr f) ive op (tr_expr e1)
			end
		    | Array_length(_) -> ill_typed "tr_statement(array_length)"
		end
	    | Assign_array(e1,e2,op,e3) -> 
		ill_typed "tr_statement(assign_array)"
	    | Typed_assign_array(t,e1,e2,op,e3) -> 
		begin
		  match op with
		  | "=" ->
		      tr_array_assignment t e1 e2 (is_pure e3) (tr_expr e3)
		  | _ -> 
		      unsupported e.expr_loc 
			("tr_statement(assign_array("^op^"))")
		end			
	    | If_expr(e1,e2,e3) -> ill_typed "tr_statement(if_expr)"
	    | Field_access(fa) -> ill_typed "tr_statement(field_access)"
(*
	    | Method_call(Static_class(ce),(_,id),args) ->
		build_why_app 
		  (Why.Var("static_method_" ^ (snd ce.class_entry_name) ^ "_" ^ id)) args
	    | Method_call(e,(_,id),args) ->
		Why.Let
		  ("temp_for_method_call",tr_expr e,
		   build_why_app 
		     (Why.make_app
			(Why.Var("method_" ^ id ^ "_parameter"),
			 Why.Var("temp_for_method_call")))
		     args)
*)
	    | Method_call(e,(_,id),args) -> 
		ill_typed "tr_statement(method_call)"
	    | Typed_method_call(e,me,args) -> 
		    let tre = tr_method_call e me args
		    in
		    if t = Null_type
		    then tre
		    else Why.Let("krak_ignore",tre,Why.Void)
	    | Array_creation(a) -> ill_typed "tr_statement(array_creation)"
	    | Array_access(e1,e2) -> ill_typed "tr_statement(array_access)"
	    | Typed_array_access(t, e1,e2) -> ill_typed "tr_statement(typed_array_access)"
	    | Cast(t,e) -> ill_typed "tr_statement(cast)"
	    | Typed_cast(t,e) -> ill_typed "tr_statement(typed_cast)"
	    | Class_instance_creation(_,_) -> ill_typed "tr_statement(new_object)"
	    | Typed_class_instance_creation(ce,args) -> 
		unsupported e.expr_loc "tr_statement(typed_new_object)"
	    | Instanceof(e,t) -> ill_typed "tr_statement(instanceof)"
	    | Typed_instanceof(e,t) -> ill_typed "tr_statement(instanceof result)"
	    | Result -> assert false (* should never happen *)
	    | Old(_) -> assert false (* should never happen *)
	    | Fresh(_) -> assert false (* should never happen *)
	    | Quantifier(_) -> assert false (* should never happen *)
	    | Typed_quantifier(_) -> assert false (* should never happen *)
	end
    | Var_decl(v) -> ill_typed "tr_statement(var_decl)"
    | Typed_var_decl(v,init,i) -> 
	tr_local_var_declaration Location.dummy v init (tr_statement i)
    | Return(e) -> 
	begin
	  match e with
	    | None -> Why.Void (* return; *) 
	    | Some e -> tr_expr e 
	end
    | Break(l) -> 
	if l="" then
	  Why.Raise("Break",Why.Cte(Why.Prim_int(0)))
	else
	  unsupported Location.dummy "tr_statement(break label)"
	    (*i Why.Raise("Break",Why.Var("krak_label_" ^ l)) i*)
    | Continue(l) -> unsupported Location.dummy "tr_statement(continue)"
    | Label(l,s) -> unsupported Location.dummy "tr_statement(label)"
    | If_statement(e,s1,s2) ->
	Why.If(tr_expr e, tr_statement s1, tr_statement s2)
    | While(l,e,s) ->
	incr label_counter;
	let new_label = "krak_loop" ^ (string_of_int !label_counter) in
	let (inv,var) = tr_loop_inv new_label l e.expr_loc in
	let tr = Why.make_while (tr_expr e) inv var (tr_statement s) in
	let eff = Effects.statement s in
	let tr =
	  if eff.breaks then
	    Why.Try(tr,"Break","krak_label",Why.Void)
	  else
	    tr
	in
	Why.make_label new_label tr
    | For(l,el1,e,el2,s) -> 
	incr label_counter;
	let new_label = "krak_loop" ^ (string_of_int !label_counter) in
	let (inv,var) = tr_loop_inv new_label l e.expr_loc in
	let tr =
	  Why.append
	    (tr_block el1)
	    (Why.make_while (tr_expr e) inv var 
	       (tr_block (s::el2)))
	in
	let eff = Effects.statement s in
	let tr = 
	  if eff.breaks then
	    Why.Try(tr,"Break","krak_label",Why.Void)
	  else
	    tr
	in
	Why.make_label new_label tr
    | For_decl(_) -> ill_typed "tr_statement(for_decl)"
    | For_typed_decl(lve,lvi,l,e,el2,s) -> 
	incr label_counter;
	let new_label = "krak_loop" ^ (string_of_int !label_counter) in
	let (inv,var) = tr_loop_inv new_label l e.expr_loc in
	let tr = 
	  tr_local_var_declaration Location.dummy lve lvi
	    (Why.make_while (tr_expr e) inv var 
	       (tr_block (s::el2)))
	in
	let eff = Effects.statement s in
	let tr =
	  if eff.breaks then
	    Why.Try(tr,"Break","krak_label",Why.Void)
	  else
	    tr
	in
	Why.make_label new_label tr
    | Switch(e,cases) -> 
	begin
	  try
	    let (l,b) = well_formed_switch cases in
	    Why.Let("temp_switch",
		    tr_expr e,
		    List.fold_right
		      (fun (c,b) acc -> Why.If(c,tr_block b,acc))
		      l (tr_block b))
	  with
	      Exit -> 
		unsupported e.expr_loc 
		  ("tr_statement(switch not well formed)") 
	end
    | Block(s) -> tr_block s
    | Synchronized (e, _) -> 
	unsupported e.expr_loc "tr_statement(synchronized)"
    | Try (_, _, _) -> unsupported Location.dummy "tr_statement(try)"
    | Do (inv,_, e) -> unsupported e.expr_loc "tr_statement(do)"
    | Throw(e) ->  
	Why.Raise("Exception",tr_expr e)
    | Jml_assert(e) ->
	Why.Assert(tr_assertion "@init" e)
    | Annotated _ ->
	assert false
    | Typed_annotated(spec,st) ->
	Why.Triple
	  (tr_assertion "@init" spec.typed_compact_requires,
	   tr_statement st,
	   Why.make_and
	     (tr_assertion "@" spec.typed_compact_ensures)
	     (Why.make_and 
		(tr_assignable_base (Why.LVar("nothing_loc")) Location.dummy 
		   "@" spec.typed_compact_assignable) 
		!ref_modifiable),
           tr_signals_clause spec.typed_compact_signals)

and tr_begin_end i l =
  Why.Block(List.map (tr_statement) (i::l))

and tr_block (l : statement list) : Why.expr = 
  match l with
    | [] -> Why.Void
    | [i] -> tr_statement i
    | i::l -> 
	match i with
	  | Skip -> tr_block l
	  | If_statement(e,s1,s2) ->
	      begin
		try
		  match (abrupt_termination s1,abrupt_termination s2) with
		    | (false,false) -> 
			tr_begin_end i l
		    | (true,true) ->
			Format.printf "WARNING : unreachable statement";
			Why.If
			  (tr_expr e,
			   tr_statement s1,tr_statement s2)
		    | (false,true) ->
			Why.If
			  (tr_expr e,
			   tr_block (s1::l), tr_statement s2)
		    | (true,false) ->
			Why.If
			  (tr_expr e,
			   tr_statement s1,tr_block (s2::l))
		with
		    Not_found ->
			Why.If
			  (tr_expr e,
			   tr_statement s1,tr_statement s2)
	      end
	  | Typed_var_decl(lve,init,i) -> 
	      tr_statement (Typed_var_decl(lve,init,Block(i::l)))
	      (*
	      Format.printf "***Assert false in tr_block: %a@."
		(Format_misc.fprintf_list Ast.fprintf_statement) (i::l);
	      assert false
	      *)
	  | _ -> tr_begin_end i l
   
;;
  
let tr_parameter p = 
  match p with
    | Local_variable_entry le ->
	let t = tr_type le.local_variable_entry_type in
	(le.local_variable_entry_name,t)
    | _ -> ill_typed "tr_parameter(non_local)"
;;

let rec params_type params return_type = 
  match params with
    | [] -> return_type
    | (id,t)::l ->
	Why.Prod_type(id,t,params_type l return_type)
;;

let arg_pre x class_name = 
  let inv = 
    Why.LPred(class_name ^ "_invariant", 
	      [Why.LVar("heap");Why.LVar(x)])
  in
   Why.make_and
    (Why.LPred("well_formed",
	       [Why.LVar("heap");Why.LVar(x);
		Why.LApp("ClassType",[Why.LVar(class_name)])]
	      ))
    (if x="this" then inv else
       (Why.LImpl
	  (Why.LPred("ne_value",[Why.LVar(x);Why.LVar("Null")]),inv)))
    
;;

let rec args_pre p l =
  match l with
    | [] -> p
    | (x,xe)::r ->
	let t = 
	  match xe with
	    | Local_variable_entry le -> le.local_variable_entry_type
	    | _ -> ill_typed "args_pre(nonlocal)"
	in
	match t with
	  | Class_type ce ->
	      args_pre 
		(Why.make_and p (arg_pre x ce.class_entry_name))
		r
	  | Array_type _ ->
	      args_pre
		(Why.make_and 
		   p 
		   (Why.LPred("well_formed",
			      [Why.LVar("heap");
			       Why.LVar(x);
			       tr_type_term t])))
		r
	  | _ -> args_pre p r



(* Store the modifiable clause of method for automatic insertion in 
   loop invariant *)

let store_tr_assignable_base base loc l = 
  ref_modifiable := tr_assignable_base base loc "@init" l;
  tr_assignable_base base loc "@" l

let store_tr_assignable = 
    store_tr_assignable_base (Why.LVar("nothing_loc"))
;;

let reset_tr_assignable () = ref_modifiable := Why.LTrue

let tr_effects e =
  let reads =
    Inttagset.fold
      (fun ive acc ->
	 if ive.instance_variable_entry_is_static
	 then ive.instance_variable_entry_trans_name :: acc
	 else acc)
      e.reads
      (if e.reads_heap then ["heap"] else [])
  and writes = 
    Inttagset.fold
      (fun ive acc ->
	 if ive.instance_variable_entry_is_static
	 then ive.instance_variable_entry_trans_name :: acc
	 else acc)
      e.writes
      (if e.writes_heap then ["heap"] else [])
  in
  (reads,writes)

let tr_method_declaration spec_only class_name class_invariant jml me body =
  reset_tr_assignable ();
  let routine = me.method_entry_routine in
  let l = routine.routine_entry_parameters in
  let params =
    if me.method_entry_is_static
    then
      if l=[]
      then ["tt",Why.unit_type]
      else List.map (fun (_,pe) -> tr_parameter pe) l
    else
      ("this",Why.base_type("value"))::
      (List.map (fun (_,pe) -> tr_parameter pe) l)
  in
  let (jml_pre,jml_assignable,jml_post,jml_signals) = 
    match jml with
      | Typed_compact tcc ->
	  let requires = tr_assertion "" tcc.typed_compact_requires
	  and assignable = tcc.typed_compact_assignable
	  and ensures = tcc.typed_compact_ensures
	  and signals = tr_signals_clause tcc.typed_compact_signals
	  in      
	    (requires, assignable, ensures, signals)
      | _ -> ill_typed "tr_method_declaration(jml desugaring)"
  in 
  let (pre,post,signals) =    
    if me.method_entry_is_static
    then (args_pre jml_pre l,tr_assertion "@" jml_post,jml_signals)
    else ((args_pre 
	     (Why.make_and jml_pre 
		(Why.make_and 
		   (Why.LPred("ne_value",[Why.LVar "this";Why.LVar "Null"]))
		   (arg_pre "this" class_name))) l),
	  (if me.method_entry_is_pure then tr_assertion "@" jml_post else
	     Why.make_and (tr_assertion "@" jml_post) 
	       (Why.LPred(class_name ^ "_invariant", 
			  [Why.LVar("heap");Why.LVar("this")]))),
	  (if me.method_entry_is_pure then jml_signals else
	       Option.map 
		 (fun (e,p) -> 
		    (e,Why.make_and p  
		       (Why.LPred(class_name ^ "_invariant", 
				  [Why.LVar("heap");Why.LVar("this")]))))
		 jml_signals))
  in
  let assignable =
    if routine.routine_entry_effects.writes_heap then 
      store_tr_assignable Location.dummy jml_assignable
    else Why.LTrue
  in
  let post = Why.make_and post assignable
  in
  let return_type = tr_type me.method_entry_return_type in
  let (reads,writes) = tr_effects routine.routine_entry_effects in
(*i
  Format.fprintf Config.log "Effects: method %s writes" me.method_entry_name;
  List.iter (fun s -> Format.fprintf Config.log " %s" s) writes;
  Format.fprintf Config.log ".@.";
i*)
  let signals =
    match (signals,routine.routine_entry_effects.throws) with
      | (Some _,true) -> 
	  Option.map (fun (s,a) -> (s,Why.make_and a assignable)) signals
      | (None,false) -> None
      | (Some _,false) ->
	  Format.printf "Warning: method %s body cannot throw an exception,@ whereas its specification allows it.@." me.method_entry_name;
	  None
      | (None,true) ->
	  (*
	    Format.printf "Warning: method %s may throw an exception, whereas its specification does not allows it.@." me.method_entry_name;
	  *)
	  Some ("Exception",Why.LFalse)
  in  
  let why_type = params_type params return_type 
  and why_annot_type = 
    params_type params 
      (Why.Annot_type(pre,return_type,
		      reads,writes,
		      post,signals))
  in
  let why_decls =
    let extern_decl =
      Why.Param (* External *)
	(class_name ^ "_" ^ me.method_entry_name ^ "_parameter", why_annot_type)
    in
    if me.method_entry_is_pure 
    then 
      if me.method_entry_is_predicate then
	let logic_decl =
	  let t = 
	    if routine.routine_entry_effects.reads_heap then  
	      Why.Prod_type
		("",Why.base_type("store"),
		 params_type params (Why.base_type "prop"))    
	    else
	      params_type params (Why.base_type "prop")
	  in
	  Why.Logic(class_name ^ "_" ^ me.method_entry_name, t)
	in [logic_decl;extern_decl]
	     (*i
	       let logic_bool_decl =
	       Why.Logic
	       (class_name ^ "_" ^ me.method_entry_name ^ "_bool" , 
	       Why.Prod_type("",Why.Base_type("store"),why_type))	  
	       and logic_prop_decl =
	       Why.Logic
	       (class_name ^ "_" ^ me.method_entry_name, 
	       Why.Prod_type("",Why.Base_type("store"),
	       params_type params (Why.Base_type "prop")))	  
	       in [logic_prop_decl;logic_bool_decl;extern_decl]
	       i*)
      else
	let logic_decl =
	  let t = 
	    if routine.routine_entry_effects.reads_heap then
	      Why.Prod_type("",Why.base_type("store"),why_type)
	    else why_type
	  in
	  Why.Logic(class_name ^ "_" ^ me.method_entry_name,t)
	in [logic_decl;extern_decl]
    else
      [extern_decl]
  and coq_decls =
    if me.method_entry_is_pure 
    then 
      if me.method_entry_is_predicate then
	let jml_post =
	  match jml with
	    | Typed_compact tcc ->
		let ensures = tcc.typed_compact_ensures
		in      
		begin
		  match ensures.expr_node with
		    | Bin(_,"<==>",e) -> e
		    | _ -> assert false
		end
	    | _ -> assert false
	in 	  
	let prop_decl =
	  let params = 
	    if routine.routine_entry_effects.reads_heap then
	      ("heap",Why.base_type("store"))::params
	    else params
	  in
	  Why.PropDefinition
	    (class_name ^ "_" ^ me.method_entry_name, 
	     params,
	     (*i Why.make_and pre i*) (tr_assertion "@" jml_post))
	in
	[prop_decl]
	(*i
	  let parameter_decl =
	  Why.Parameter
	    (class_name ^ "_" ^ me.method_entry_name, 
	     Why.Prod_type("",Why.Base_type("store"),
			   params_type params (Why.Base_type "Prop")))
	and axiom_decl =
	  let coq_type =
	    params_type params 
	      (Why.Annot_type
		 (pre,return_type,[],[],
		  Why.LLet
		    ("result",
		     Why.LApp 
		       (class_name ^ "_" ^ me.method_entry_name,
			(Why.LVar "heap"::
			   (List.map (fun (x,_) -> Why.LVar x)) params)),
		       post),None))
	  in
	  Why.Axiom
	    (class_name ^ "_" ^ me.method_entry_name ^ "_axiom", 
	     Why.Prod_type("heap",Why.Base_type("store"),coq_type))
	and parameter_bool_decl =
	  Why.Parameter
	    (class_name ^ "_" ^ me.method_entry_name ^ "_bool", 
	     Why.Prod_type("",Why.Base_type("store"),why_type))
	and axiom_bool_decl =
	  let coq_type = 
	    params_type params 
	      (Why.Annot_type
		 (pre,return_type,[],[],
		  Why.LLet
		    ("result",
		     Why.LApp 
		       (class_name ^ "_" ^ me.method_entry_name ^ "_bool",
			(Why.LVar "heap"::(List.map (fun (x,_) -> Why.LVar x) params))),
		       post),None))
	  in 	  
	  Why.Axiom
	    (class_name ^ "_" ^ me.method_entry_name ^ "_bool_axiom", 
	     Why.Prod_type("heap",Why.Base_type("store"),coq_type))
	in      
	[parameter_decl;axiom_decl;parameter_bool_decl;axiom_bool_decl]
      else
i*)	
      else
	let parameter_decl =
	  let t = 
	    if routine.routine_entry_effects.reads_heap then
	      Why.Prod_type("",Why.base_type("store"),why_type)
	    else why_type
	  in	  
	  Why.Parameter(class_name ^ "_" ^ me.method_entry_name, t)
	and axiom_decl =
	  let par1 =
	    if routine.routine_entry_effects.reads_heap then
	      Why.LVar "heap"::
	      (List.map (fun (x,_) -> Why.LVar x) params)
	    else
	      List.map (fun (x,_) -> Why.LVar x) params
	  in
	  let coq_type = 
	    params_type params 
	      (Why.Annot_type
		 (pre,return_type,[],[],
		  Why.LLet
		    ("result",
		     Why.LApp 
		       (class_name ^ "_" ^ me.method_entry_name,par1),
		       post),None))
	  in
	  let par2 =
	    if routine.routine_entry_effects.reads_heap then
	      Why.Prod_type("heap",Why.base_type("store"),coq_type)
	    else coq_type
	  in
	  Why.Axiom
	    (class_name ^ "_" ^ me.method_entry_name ^ "_axiom", 
	     par2)
	in      
	[parameter_decl;axiom_decl]
    else []
  in
  if spec_only then  why_decls,coq_decls
  else
    match body with
      | Some l ->
	  begin
	    try
	      ((Why.Def
		  (class_name ^ "_" ^ me.method_entry_name, 
		   Why.Fun
		     (params,pre,tr_block l,post,signals))
		::why_decls),coq_decls)
	    with
	      | Unsupported msg ->
		  Format.printf 
		    "**Warning: translation of body of %s not performed due to unsupported feature in translation: %s@."  
		    me.method_entry_name msg;
		  (why_decls,coq_decls)
	  end
      | None -> why_decls,coq_decls
;;

let tr_constr_declaration spec_only class_name class_invariant jml ce body =
  let routine = ce.constructor_entry_routine in
  let l = routine.routine_entry_parameters in
  let params = 
    (* ("this",Why.Base_type("value")):: *)
    if l=[]
    then ["tt",Why.unit_type]
    else
      List.map (fun (_,pe) -> tr_parameter pe) l
  in
  let (jml_pre,jml_assignable,jml_post,jml_signals) = 
    match jml with
      | Typed_compact tcc ->
	  let requires = tr_assertion "" tcc.typed_compact_requires
	  and assignable = tcc.typed_compact_assignable
	  and ensures = 
	    begin
	      this_is_result := true;
	      let r = tr_assertion "@" tcc.typed_compact_ensures in
	      this_is_result := false;
	      r
	    end
	  and signals = tr_signals_clause tcc.typed_compact_signals
	  in      
	    (requires, assignable, ensures, signals)
	  | _ -> ill_typed "JML Desugaring failed"
  in
  let (pre,post,signals) =    
    (args_pre 
       jml_pre
       (*
	 (Why.make_and jml_pre 
	  (Why.make_and 
	     (Why.LPred
		("well_formed",
		 [Why.LVar("heap");Why.LVar("result");
		  Why.LApp
		    ("ClassType",
		     [Why.LVar
			(ce.constructor_entry_class.class_entry_name)])]))
		(Why.LPred("ne_value",[Why.LVar "this";Why.LVar "Null"]))))
       *)
       l,
       Why.make_and jml_post 
	 (Why.make_and 
	    (Why.LPred("ne_value",[Why.LVar "result";Why.LVar "Null"]))
	    (Why.make_and 
	       (Why.LPred
		  ("well_formed",
		   [Why.LVar("heap");Why.LVar("result");
		    Why.LApp
		      ("ClassType",
		       [Why.LVar
			  (class_name)])]))
	       (Why.make_and
		  (Why.LPred("fresh",[Why.LVar("heap@");Why.LVar("result")]))
		  (Why.LPred(class_name ^ "_invariant", 
			     [Why.LVar("heap");Why.LVar("result")]))))) ,
       jml_signals)
  in
  let assignable = 
    store_tr_assignable (* _base *)
(*
      (Why.LApp("allcells_loc",[Why.LVar("this")]))
*)
      
      Location.dummy jml_assignable
  in
  let post = Why.make_and post assignable 
  and signals =
    match (signals,routine.routine_entry_effects.throws) with
      | (Some _,true) -> 
	  Option.map (fun (s,a) -> (s,Why.make_and a assignable)) signals
      | (None,false) -> None
      | (Some _,false) ->
	  Format.printf "Warning: constructor %s body cannot throw an exception,@ whereas its specification allows it.@." ce.constructor_entry_class.class_entry_name;
	  None
      | (None,true) ->
	  (*
	    Format.printf "Warning: method %s may throw an exception, whereas its specification does not allows it.@." me.method_entry_name;
	  *)
	  Some ("Exception",Why.LFalse)
  in  
  let return_type = Why.base_type("value") in
  let (reads,writes) = tr_effects routine.routine_entry_effects in 
  let why_type = params_type params return_type 
  and why_annot_type = 
    params_type params 
      (Why.Annot_type(pre,return_type,reads,writes,post,signals))
  in
  let extern_decl =
    Why.Param (* External *)
      (ce.constructor_entry_trans_name ^ "_parameter", why_annot_type) 
  in  
  if spec_only then [extern_decl], []
  else
    let tr_body = 
      (*
	let this = (new_obj heap classid) in
        heap := (store_new_ref heap new_val);
        <body>;
        this
      *)

(*i
	let new_val = "this" in
	Why.Let
	  (new_val,
	   Why.make_app "new_obj"
	     [Why.Deref("heap");
	      Why.Var(ce.constructor_entry_class.class_entry_name)],
	   Why.Block
	     ([Why.Assign
		 ("heap",
		  Why.make_app "store_new_ref"
		    [Why.Deref("heap");
		     Why.Var(new_val)]);
	       tr_block body;
	       Why.Var(new_val)]))
i*)
	let new_val = "this" in
	Why.Let
	  (new_val,
	   Why.make_app
	     "alloc_new_obj"
	     [Why.Var(ce.constructor_entry_class.class_entry_name)],
	   Why.Block
	     ([tr_block body;
	       Why.Var(new_val)]))
	  in
	  ([Why.Def(ce.constructor_entry_trans_name, 
		       Why.Fun(params,pre,tr_body,post,signals));
	    extern_decl],[])
;;

let tr_instance_variable_declaration class_name ve = 
  if !Config.verbose >=2 then
    Format.printf "producing declaration of field %s.@."
      ve.instance_variable_entry_trans_name;
  ([ Why.External
       (tr_ive ve,
	Why.base_type("fieldId"))
       ;
     Why.Logic
       (tr_ive ve, 	     
	Why.base_type("fieldId"))
   ] ,
   [])
;;


let tr_static_variable_declaration loc v init =
  let i = 
    match init with
    | None -> type_default_value false v.instance_variable_entry_type
    | Some e -> 
	type_default_value false v.instance_variable_entry_type
	(*i 
	  tr_variable_initializer loc e
	  i*)
  in
  let name = v.instance_variable_entry_trans_name
  and t = tr_type v.instance_variable_entry_type 
  in
  if v.instance_variable_entry_is_final
  then
    let coq_decl =
      match init with
	| None -> [Why.Parameter(name,t)]
	| Some vi ->
		begin
		  match vi with
		    | Simple_initializer _ ->
			
			[Why.Parameter(name,t) ;
			 Why.Axiom
			   (name ^ "_value",
			    Why.Annot_type
			      (Why.LTrue,t,[],[],
			       Why.LPred("eq",
					 [Why.LVar(name);
					  tr_term_variable_initializer 
					    Location.dummy vi]),None))
			]	      
		    | Array_initializer _ ->
			[Why.Parameter(name,t)]
		end
    in
    ([Why.Logic(name,t); Why.External(name,t)], coq_decl )
  else
    ([Why.Param(name,Why.Ref_type(t))],
     [])


let rec find_method_name m =
  match m with
    | Simple_method_declarator((loc,id),_) -> id
    | Array_method_declarator(m) -> find_method_name m

let name_of_field_declaration f =
  match f with 
    | Typed_method(me,jml,b) -> "method " ^ me.method_entry_name
    | Typed_constructor(ce,jml,eci,b) -> 
	"constructor " ^ ce.constructor_entry_class.class_entry_name
    | Typed_variable(ve,init) -> "field " ^ ve.instance_variable_entry_name
    | Static_initializer(_) -> "static initializer"
    | Method(md,_) -> "method " ^ find_method_name md.method_declarator
    | Constructor(cd) -> "constructor " ^ (snd cd.constr_name)
    | Variable(v) -> "field"
    | Specification(_) -> "jml specification"

let tr_field_declaration class_name class_invariant f = 
  try
    match f with 
      | Typed_method(me,jml,b) -> 
	  tr_method_declaration 
	    (not me.method_entry_routine.routine_entry_produce_body)
	  class_name class_invariant jml me b
      | Typed_constructor(ce,jml,eci,b) ->
	  begin
	    match eci with
	      | Invoke_none -> 
		  tr_constr_declaration 
		  (not ce.constructor_entry_routine.routine_entry_produce_body)
		  class_name class_invariant jml ce b
	      | Invoke_this(el) -> 
		  if ce.constructor_entry_routine.routine_entry_produce_body then
		    unsupported Location.dummy
		      "tr_field_declaration(constructor_invoke(this))"
		  else
		    tr_constr_declaration true
		      class_name class_invariant jml ce b
	      | Invoke_super(el) -> 
		  if ce.constructor_entry_routine.routine_entry_produce_body then
		    unsupported Location.dummy
		      "tr_field_declaration(constructor_invoke(super))"
		  else
		    tr_constr_declaration true
		      class_name class_invariant jml ce b
	  end      
      | Typed_variable(ve,init) ->
	  if ve.instance_variable_entry_is_static
	  then tr_static_variable_declaration Location.dummy ve init
	  else tr_instance_variable_declaration class_name ve
      | Static_initializer _ -> 
	  unsupported Location.dummy "tr_field_declaration(static_init)"
      | Method(_,_) -> 
	  ill_typed "tr_field_declaration(method)"
      | Constructor(_) -> 
	  ill_typed "tr_field_declaration(constructor)"
      | Variable(_) -> 
	  ill_typed "tr_field_declaration(variable)"
      |	Specification(_) -> ([],[]) (* already computed before *)
  with
    | Unsupported msg when not !Config.stop_on_error  ->
	Format.printf 
	  "**Warning: translation of %s not performed due to unsupported feature in translation: %s@." 
	  (name_of_field_declaration f) msg;
	([],[])
    | Typing.Ill_typed msg when not !Config.stop_on_error  -> 
	Format.printf 
	  "**Warning: translation of %s not performed due to typing errors (%s)@."
	  (name_of_field_declaration f) msg;
	([],[])

let collect_class_invariant class_name class_fields =
  try
    List.fold_left
      (fun acc (doc,f) ->
	 match f with
	   | Specification s ->
	       List.fold_left
		 (fun acc s ->
		   match s with
		   | Invariant e -> Why.make_and acc (tr_assertion "" e)
		   | _ -> acc)
		 acc s
	   | _ -> acc)
      Why.LTrue
      class_fields
  with
    | Unsupported msg when not !Config.stop_on_error  ->
	Format.printf 
	  "**Warning: translation of class invariant of %s not performed due to unsupported feature in translation: %s@." 
	  class_name msg;
	Why.LTrue
      | Typing.Ill_typed msg when not !Config.stop_on_error  -> 
	  Format.printf 
	    "**Warning: translation of class invariant of %s not performed due to typing errors (%s)@."
	    class_name msg;
	  Why.LTrue	
;;

let tr_class_declaration ce c =
  let class_name = ce.class_entry_name in
  try
    let inv = collect_class_invariant class_name c.class_fields in
    let inv_name = class_name ^ "_invariant" in
    let create_decl =
(*i ??
      if List.mem `FINAL c.class_modifiers then ([],[])
      else
i*)
	([(*i
	    Why.External
	    ("create_" ^ class_name, 
	     Why.Base_type("value (*" ^ class_name ^ "*)"));
	    i*)
	  Why.External(class_name,Why.base_type("classId"));
	  Why.Logic(class_name,Why.base_type("classId"));
	  Why.Logic
	    (inv_name,
	     Why.Prod_type
	       ("",Why.base_type("store"),
		Why.Prod_type
		  ("this",
		   Why.base_type("value (*" ^ class_name ^ "*)"),
		   Why.base_type("prop")))
	    )],
	 [Why.PropDefinition
	    (inv_name,	   
	     (["heap", Why.base_type("store"); 
	       "this",Why.base_type("value")]  ),
	       inv);
           Why.Verbatim("Hint Unfold "^inv_name^".");
	 ])
    in
    List.fold_left
      (fun (wdacc,cdacc) (_,f) -> 
	 let (wd,cd) = tr_field_declaration class_name inv f
	 in (List.rev_append wd wdacc,List.rev_append cd cdacc))
      create_decl
      c.class_fields
      
  with
    | Unsupported msg when not !Config.stop_on_error  ->
	Format.printf 
	  "**Warning: translation of class %s not performed due to unsupported feature in translation: %s@." 
	  class_name msg;
	([],[])
    | Typing.Ill_typed msg when not !Config.stop_on_error  -> 
	Format.printf 
	  "**Warning: translation of class %s not performed due to typing errors (%s)@."
	  class_name msg;
	([],[])	
;;

let tr_interface_member_declaration interface_name interface_invariant f =
  try
    match f with
    | Constant_declaration(_) -> 
	ill_typed "tr_interface_member_declaration(constant)"
    | Typed_constant_declaration(ive,i) ->
	let name = interface_name ^ "_" ^ ive.constant_entry_name 
	and t = tr_type ive.constant_entry_type
	in
	let coq_decl =
	  match i with
	    | None -> [Why.Parameter(name,t)]
	    | Some vi ->
		begin
		  match vi with
		    | Simple_initializer _ ->
			[Why.Parameter(name,t) ;
			 Why.Axiom
			   (name ^ "_value", 
			    Why.Annot_type
			      (Why.LTrue,t,[],[],
			       Why.LPred("eq",
					 [Why.LVar(name);
					  tr_term_variable_initializer 
					    Location.dummy vi]),None))
			]
		    | Array_initializer _ -> 
			[Why.Parameter(name,t)]
		end
	in
	([Why.Logic(name,t);Why.External(name,t)],coq_decl)
    | Method_header(_) -> 
	ill_typed "tr_interface_member_declaration(method)"
    | Typed_method_header(me,jml) -> 
	([],[])
	(*i tr_method_declaration true
	  interface_name interface_invariant jml me None i*)
  with
  | Unsupported msg when not !Config.stop_on_error ->
      Format.printf 
	"**Warning: translation of interface member %s not performed due to unsupported feature in translation: %s@." 
	"(?)" msg;
      ([],[])
  | Typing.Ill_typed msg when not !Config.stop_on_error  -> 
      Format.printf 
	"**Warning: translation of interface member %s not performed due to typing errors (%s)@."
	"(?)" msg;
      ([],[])	
;;

let tr_interface_declaration i =
  List.fold_left
    (fun (wdacc,cdacc) (_,f) -> 
       let (wd,cd) = tr_interface_member_declaration 
		       (snd i.interface_name) (Lit(Bool(true))) f
       in (List.rev_append wd wdacc,List.rev_append cd cdacc))
    ([],[])
    i.interface_members
;;

let tr_type_declaration d = 
  match d with
  | Interface _ -> assert false
  | Class _ -> assert false
  | Typed_interface(_, i) -> tr_interface_declaration i
  | Typed_class(ce,c) -> tr_class_declaration ce c
;;

let tr_compilation_unit a = 
  List.fold_left
    (fun (wdacc,cdacc) (_,d) -> 
       let (wd,cd) = tr_type_declaration d
       in (List.rev_append wd wdacc,List.rev_append cd cdacc)) 
    ([],[])
    a.cu_type_decls;;

(*
Local Variables: 
mode: tuareg
compile-command: "make"
End: 
*)

